#include <stdlib.h>
#include <stdio.h>
#include "config.h"

#ifdef TDF_HP_COMPILE
#define _INCLUDE_HPUX_SOURCE
#define _INCLUDE_POSIX_SOURCE
#include "HP_types.h" /* This is a hacked version of sys/types.h */
#include <rpc/types.h>
#include <rpc/xdr.h>
#undef _INCLUDE_HPUX_SOURCE
#undef _INCLUDE_POSIX_SOURCE

#else
#include <rpc/types.h>
#include <rpc/xdr.h>
#endif

#include "tdf.h"

static void *tdf_get_spc(int num, size_t size);


static void  write_int_array(
  int *pt,        /* array of ints */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  );

static void  write_float_array(
  float *pt,        /* array of floats */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  );

static void  read_int_array(
  int **pt,         /* array of ints */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  );

static void  read_float_array(
  float **pt,       /* array of floats */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  );




void tdf_init_xform(struct system_matrix *xform,
  int type,           /* =1 => forward projection                       */
                      /* =0 => back projection                          */
  int n_x,            /* number of samples in x image coordinate        */
  int n_y,            /* number of samples in y image coordinate        */
                      /* Sparse matrix column is computed by:           */
                      /* j = m*n_x+n where m  and n are respectively    */
                      /* the y and x indices of the pixels location.    */
  int Mpro,           /* number of projections                          */
  int NN              /* NN specifies the total number of nonzero       */
                      /* elements in the sparse matrix.                 */
  )
/* This subroutine initializes variables and allocates memory.          */
/* HOWEVER, IT DOES NOT INITIALIZE THE ARRAY OF POINTERS IN THE SPARSE  */
/* MATRIX.                                                              */
{
  int n;

  if( (type!=FORWARD)&&(type!=BACKWARD) ) {
      fprintf(stderr,"\nERROR:tdf_init_xform - invalid type\n");
      exit(-1);
    }

  xform->type = type;
  xform->n_x = n_x;
  xform->n_y = n_y;
  xform->Npix = n_x*n_y;
  xform->Mpro = Mpro;

  if(xform->type==FORWARD) {
    n = xform->Npix;
  }
  else {
    n = xform->Mpro;
  }

  xform->n_row 
    = (int *) tdf_get_spc(n, sizeof(int));

  /* Allocate array of pointers */
  xform->val = (float **)tdf_get_spc(n, sizeof(float *));
  xform->row = (int **)tdf_get_spc(n, sizeof(int *));

  /* Allocate block of memory */
  xform->val[0]  = (float *)tdf_get_spc(NN, sizeof(float));
  xform->row[0]  = (int *)tdf_get_spc(NN, sizeof(int));
}


void tdf_set_img_geom(struct system_matrix *xform,
  int    used,          /* 1 => used; 0 => not used */
  double delta_x,       /* x = (i-x_0)*delta_x */
  double x_0,
  double delta_y,       /* y = (i-y_0)*delta_y */
  double y_0
  )
{
  xform->img_geom.used = used;

  xform->img_geom.delta_x = delta_x;
  xform->img_geom.x_0 = x_0;
  xform->img_geom.delta_y = delta_y;
  xform->img_geom.y_0 = y_0;
}




int tdf_write_system_matrix(struct system_matrix *xform, 
  char *fname     /* character string containing file name */
  )
/* Writes out system_matrix data structure. */
/* Returns 1 if successful; 0 otherwise. */
{
  XDR xdr;
  FILE  *fout;
  int magik_cookie;
  int used = USED, unused = UNUSED;
  int n,NN,i;

  fout = fopen(fname, "w");
  if (fout == NULL) {
    fprintf(stderr,"Cannot write to file named %s\n",fname);
    return(0);
  }
  xdrstdio_create(&xdr, fout, XDR_ENCODE);

  /* write out file type identifier */
  magik_cookie=TDF_IDENTIFIER+1;
  xdr_int(&xdr,&magik_cookie );

  /* write out xform structure */
  xdr_int(&xdr,&(xform->type) );
  xdr_int(&xdr,&(xform->Npix) );
  xdr_int(&xdr,&(xform->Mpro) );
  xdr_int(&xdr,&(xform->n_x) );
  xdr_int(&xdr,&(xform->n_y) );

  /* Write out sparse matrix */
  if(xform->type==FORWARD) {
    n = xform->Npix;
  }
  else 
    if(xform->type==BACKWARD) {
      n = xform->Mpro;
    }
    else { 
      fprintf(stderr,"\nERROR:tdf_write_system_matrix - system matrix type ");
      fprintf(stderr,"unspecified\n");
      exit(-1);
    }

  write_int_array(xform->n_row, n, &xdr);

  NN = 0;
  for(i=0; i<n; i++) NN += xform->n_row[i];

  write_float_array(xform->val[0], NN, &xdr);
  write_int_array(xform->row[0], NN, &xdr);

  /* Write out image_geometry structure */
  if (xform->img_geom.used == USED) {
    xdr_int(&xdr,&used );
    xdr_double(&xdr,&(xform->img_geom.delta_x) );
    xdr_double(&xdr,&(xform->img_geom.x_0) );
    xdr_double(&xdr,&(xform->img_geom.delta_y) );
    xdr_double(&xdr,&(xform->img_geom.y_0) );
  }
  else {
    xdr_int(&xdr,&unused );
  }

  XDR_DESTROY(&xdr);
  fclose(fout);
  return(1);
}


static void  write_int_array(
  int *pt,          /* array of ints */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  )
{
  int used = USED, unused = UNUSED;
  int i;

  if(pt==NULL) {
    xdr_int(xdr, &unused );
  }
  else {
    xdr_int(xdr, &used );
    for(i=0; i<length; i++) xdr_int(xdr, pt+i );
  }
}



static void  write_float_array(
  float *pt,        /* array of floats */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  )
{
  int used = USED, unused = UNUSED;
  int i;

  if(pt==NULL) {
    xdr_int(xdr, &unused );
  }
  else {
    xdr_int(xdr, &used );
    for(i=0; i<length; i++) xdr_float(xdr, pt+i );
  }
}




int tdf_read_system_matrix(struct system_matrix *xform, 
  char *fname     /* character string containing file name */
  )
/* Reads data structure. Returns 1 if successful; 0 otherwise. */
{
  XDR   xdr;
  FILE  *fin;
  int magik_cookie;
  int used;
  int n;
  int NN,i;

  fin = fopen(fname, "r");
  if (fin == NULL) {
    fprintf(stderr,"Cannot read file named %s\n",fname);
    exit (-1);
  }
  xdrstdio_create(&xdr, fin, XDR_DECODE);

  /* read file type identifier */
  xdr_int(&xdr, &magik_cookie );
  if(magik_cookie!=(TDF_IDENTIFIER+1)) {
    fprintf(stderr,"\nError: File %s is not an etdf file type\n",fname);
    return(0);
  }

  xdr_int(&xdr, &(xform->type) );
  xdr_int(&xdr, &(xform->Npix) );
  xdr_int(&xdr, &(xform->Mpro) );
  xdr_int(&xdr, &(xform->n_x) );
  xdr_int(&xdr, &(xform->n_y) );

  /* read sparse matrix */
  if(xform->type==FORWARD) {
    n = xform->Npix;
  }
  else 
    if(xform->type==BACKWARD) {
      n = xform->Mpro;
    }
    else { 
      fprintf(stderr,"\nERROR:tdf_read_system_matrix - system matrix type ");
      fprintf(stderr,"unspecified\n");
      exit(-1);
    }

  read_int_array( &(xform->n_row), n, &xdr );

  NN = 0;
  for(i=0; i<n; i++) NN += xform->n_row[i];

  xform->val = (float **)tdf_get_spc(n,sizeof(float *));
  xform->row = (int **)tdf_get_spc(n,sizeof(int *));

  read_float_array( &(xform->val[0]), NN, &xdr);
  read_int_array( &(xform->row[0]), NN, &xdr);

  for(i=0; i<n-1; i++) {
    xform->val[i+1] = xform->val[i] + xform->n_row[i];
    xform->row[i+1] = xform->row[i] + xform->n_row[i];
  }

  /* read image_geometry structure */
  xdr_int(&xdr, &used );
  if(used==USED) {
    xform->img_geom.used = USED;
    xdr_double(&xdr, &(xform->img_geom.delta_x) );
    xdr_double(&xdr, &(xform->img_geom.x_0) );
    xdr_double(&xdr, &(xform->img_geom.delta_y) );
    xdr_double(&xdr, &(xform->img_geom.y_0) );
  }
  else {
    xform->img_geom.used = UNUSED;
  }

  XDR_DESTROY(&xdr);
  fclose(fin);
  return(1);
}



static void  read_int_array(
  int **pt,         /* array of ints */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  )
{
  int used;
  int i;

  xdr_int(xdr, &used );

  if(used==USED) {
    *pt = (int *)tdf_get_spc(length,sizeof(int));
    for(i=0; i<length; i++) xdr_int(xdr, *pt+i);
  }
  else {
    *pt = NULL;
  }
}


static void  read_float_array(
  float **pt,       /* array of floats */
  int length,       /* length of array */
  XDR *xdr          /* pointer to file */
  )
{
  int used;
  int i;

  xdr_int(xdr, &used );

  if(used==USED) {
    *pt = (float *)tdf_get_spc(length,sizeof(float));
    for(i=0; i<length; i++) xdr_float(xdr, *pt+i);
  }
  else {
    *pt = NULL;
  }
}



void destroy_system_matix(struct system_matrix *xform)
{
  if(xform->n_row!=NULL) {
    free((void *)xform->n_row);
  }
  if(xform->val[0]!=NULL) {
    free((void *)xform->val[0]);
    free((void *)xform->row[0]);
  }
  if(xform->val!=NULL) {
    free((void *)xform->val);
    free((void *)xform->row);
  }
}











static void *tdf_get_spc(int num, size_t size)
{
	void *pt;

	if( (pt=calloc((size_t)num,size)) == NULL ) {
		fprintf(stderr, "\nERROR: tdf_get_spc ==> calloc() error\n");
		exit(-1);
		}
	return(pt);
}
